/*
 * Copyright (C) 2010  Ikanos Communications. All rights reserved.
 * The information and source code contained herein is the property
 * of Ikanos Communications.
 */

/*
 *  BRIEF MODULE DESCRIPTION: VX185 interrupt routines.
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/irq.h>
#include <linux/kernel_stat.h>
#include <linux/module.h>
#include <linux/signal.h>
#include <linux/bitops.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/delay.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/mipsregs.h>
#include <asm/system.h>

#include <linux/kernel_stat.h>

#include <vx185_int.h>

#undef DEBUG_IRQ
#ifdef DEBUG_IRQ
/* note: prints function name for you */
#define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define ALL_INTS (IE_IRQ0 | IE_IRQ1 | IE_IRQ2 | IE_IRQ3 | IE_IRQ4 | IE_IRQ5)

/* 
 * IPC Source interrupt map table for all 64 source interrupts in the following format:
 * {
 *  Reserved bit, Trigger_type (0: Level, 1: Edge), Sync_enable(0: Bypass, 1: Enable),
 *  Polarity (0: Active low, 1: Active high), IPC Destination line  (0..13)
 *  }
 * Note that GPT0 should always be mapped to IPC line 6 (Host mips interrupt line 5)
 * since the low level assembly code hardwires CAUSEF_IP7 to go to the timer ISR.
 */

static const ipc_src_map_t ipc_src_table[64] = {
        /* 0 - 7 */
 	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 1, FUSIV_IPC_DEST_PIT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},

	/* 8 - 15 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_MAILBOX},   /* Mailbox Interrupt */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* GPT0 */
	{0, 1, 0, 1, FUSIV_IPC_DEST_GPT1},	/* GPT1 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* GPT2 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* GPT3 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},

	/* 16 - 23 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_DSP0FL0},	/* DSP0 FL0 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_DSP0FL1},	/* DSP0 FL1 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* DSP0 PF0 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* DSP0 FP1 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_DSP1FL0},	/* DSP1 FL0 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_DSP1FL1},	/* DSP1 FL1 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* DSP1 PF0 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},	/* DSP1 PF1 */

	/* 24 - 31 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 1, FUSIV_IPC_DEST_SERIAL},	/* UART0 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 1, FUSIV_IPC_DEST_SATA},      /* SATA */

	/* 32 - 39 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_PCIE0},	/* PCIE 0 (priority may be changed later) */
	{0, 0, 0, 1, FUSIV_IPC_DEST_PCIE1},	/* PCIE 1 (priority may be changed later) */
	{0, 0, 0, 0, FUSIV_IPC_DEST_USBO},	/* USB - OHCI */
	{0, 0, 0, 1, FUSIV_IPC_DEST_USBE},	/* USB - EHCI */
	{0, 0, 0, 1, FUSIV_IPC_DEST_GIGE0},
	{0, 0, 0, 1, FUSIV_IPC_DEST_GIGE1},
	{0, 0, 0, 1, FUSIV_IPC_DEST_VAP},
	{0, 0, 0, 1, FUSIV_IPC_DEST_CLSAP1},

	/* 40 - 47 */
	{0, 0, 0, 1, FUSIV_IPC_DEST_CLSAP2},
	{0, 0, 0, 1, FUSIV_IPC_DEST_BMU},
	{0, 0, 0, 1, FUSIV_IPC_DEST_SPA},
	{0, 0, 0, 1, FUSIV_IPC_DEST_EAP},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},

	/* 48 - 55 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},

	/* 56 - 63 */
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT},
	{0, 0, 0, 0, FUSIV_IPC_DEST_DEFAULT}
};

volatile ipc_t *ipc = (volatile ipc_t *)VX185_IPC_BASE_ADDR;

#ifdef CONFIG_KGDB
extern void breakpoint(void);
#endif
#if defined (CONFIG_CPU_MIPSR2_IRQ_VI)
extern void __init mips_cpu_irq_init(void);
#endif
extern void set_debug_traps(void);
extern irq_cpustat_t irq_stat [NR_CPUS];
unsigned int local_bh_count[NR_CPUS];
unsigned int local_irq_count[NR_CPUS];

void startup_ipc_src(unsigned int irq_nr);
void shutdown_ipc_src(unsigned int irq_nr);
static inline void end_ipc_src(unsigned int irq_nr);
static inline void enable_ipc_src(unsigned int irq_nr);
static inline void disable_ipc_src(unsigned int irq_nr);
void ack_ipc_src(unsigned int irq_nr);
static inline void mask_ipc_src(unsigned int irq_nr);
static inline void mask_ack_ipc_src(unsigned int irq_nr);
static inline void unmask_ipc_src(unsigned int irq_nr);
static inline void eoi_ipc_src(unsigned int irq_nr);
void configure_ipc_src (unsigned int irq_nr);

/*________________________________SUPPORTING ROUTINES_______________________________*/

static inline void enable_ipc_src(unsigned int irq_nr)
{
	unsigned char ipc_reg_select = irq_nr/8, ipc_src = irq_nr % 8;
	unsigned int src_enable = ipc->src_ctrl[ipc_reg_select];

	/* Enable the source */
	src_enable |= (IPC_SRC_ENABLE << (ipc_src * 4));
	ipc->src_ctrl[ipc_reg_select] = src_enable;

	/* Ensure the write before we go back */
//	wbflush();
}

static inline void disable_ipc_src(unsigned int irq_nr)
{
	unsigned char ipc_reg_select = irq_nr/8, ipc_src = irq_nr % 8;
	unsigned int src_enable = ipc->src_ctrl[ipc_reg_select];

	/* Disable the source */
	src_enable &= ~(IPC_SRC_ENABLE << (ipc_src * 4));
	ipc->src_ctrl[ipc_reg_select] = src_enable;

	/* Ensure the write before we go back */
//	wbflush();
}

/*
 * NOTE: This is only needed for Edge detected source interrupts
 * to ack the IRQ in IPC if its configured as Edge detect 
 */
void ack_ipc_src(unsigned int irq_nr)
{
	unsigned char dest_line = ipc_src_table[irq_nr].ipc_dst_line;
	unsigned int int_status = ipc->status_low[dest_line];
	
	if (irq_nr < 32)
		int_status |= (0x1 << irq_nr);
	else
		int_status |= (0x1 << (irq_nr-32));
	
	ipc->status_low[dest_line] = int_status;
}


static inline void mask_ipc_src(unsigned int irq_nr)
{
        /* We have already done it from our dispatcher code so
           nothing to do when kernel calls this . */
        //local_disable_irq(irq_nr);
}

static inline void mask_ack_ipc_src(unsigned int irq_nr)
{
        /* We have already done it from our dispatcher code so
           nothing to do when kernel calls this . */
        //local_disable_irq(irq_nr);
}

static inline void unmask_ipc_src(unsigned int irq_nr)
{
        /* We have already done it from our dispatcher code so
           nothing to do when kernel calls this . */
        //local_disable_irq(irq_nr);
}

static inline void eoi_ipc_src(unsigned int irq_nr)
{
        /* We have already done it from our dispatcher code so
           nothing to do when kernel calls this . */
        //local_disable_irq(irq_nr);
}

static inline void end_ipc_src(unsigned int irq_nr)
{
	if (!(irq_desc[irq_nr].status & (IRQ_DISABLED|IRQ_INPROGRESS)))
		enable_ipc_src(irq_nr);
}

void startup_ipc_src(unsigned int irq_nr)
{
	/* Configure source interrupt */
        configure_ipc_src (irq_nr);

	/* Enable the IRQ */
        enable_ipc_src(irq_nr);
}
EXPORT_SYMBOL(startup_ipc_src);

void shutdown_ipc_src(unsigned int irq_nr)
{
	disable_ipc_src(irq_nr);
	return;
}
EXPORT_SYMBOL(shutdown_ipc_src);

/* Set the following for the source interrupt:-
	1.) trigger type  - edge or level
	2.) polarity - active-low or active-high detect
	3.) source - IPC sync bypass or enable 
	4.) Destination Interrupt line (0..13)
*/
void configure_ipc_src(unsigned int irq_nr) {

        unsigned char ipc_reg_select = irq_nr/8, ipc_src = irq_nr % 8;
	unsigned int src_enable = ipc->src_ctrl[ipc_reg_select];
	unsigned int dest_enable = ipc->dest_select[ipc_reg_select];
	
	src_enable &= ~(IPC_SRC_CTRL_MASK << (ipc_src * 4));

	/* Edge or level detect */
	if (ipc_src_table[irq_nr].trigger_type) {
		src_enable |= (IPC_SRC_EDGE_DETECT << (ipc_src * 4));
	}
	
	/* Sync enable or bypass */
	if (ipc_src_table[irq_nr].sync_enable) {
		src_enable |= (IPC_SRC_SYNC_ENABLE << (ipc_src * 4));
	}

	/* Active high or low detect */
	if (ipc_src_table[irq_nr].polarity) {
		src_enable |= (IPC_SRC_ACT_HI_DETECT << (ipc_src * 4));
	}

	ipc->src_ctrl[ipc_reg_select] = src_enable;
	
	/* Configure destination line */
	dest_enable &= ~(IPC_DEFAULT_DEST_LINE << (ipc_src * 4));
	dest_enable |= (ipc_src_table[irq_nr].ipc_dst_line << (ipc_src * 4));

	ipc->dest_select[ipc_reg_select] = dest_enable;
}

static struct hw_interrupt_type vx185_ipc = {
	"Fusiv VX185",
        startup_ipc_src,
        shutdown_ipc_src,
        enable_ipc_src,
        disable_ipc_src,
        ack_ipc_src,
        mask_ipc_src, 
        mask_ack_ipc_src,
        unmask_ipc_src,
        eoi_ipc_src,
        end_ipc_src,
	NULL,
	NULL,
	NULL,
	NULL
};

void fusiv_dispatch_irq(int prio, struct pt_regs *regs)
{
	/* Let us put some memory rather then few clocks. Memory is always 
	   cheaper then CPU time . */
	unsigned char IRQCnt1  = 0;
	unsigned char IRQCnt2  = 0;
	unsigned char i;
	unsigned char IRQSet1[32];
	unsigned char IRQSet2[32];
	unsigned long stat;
	int irq;

	/* 
	 * Check if the interrupt line is owned by the periodic timer.
	 * We dont expect this line to be shared.
	 */
	if (prio == PERIODIC_TIMER_INT_PRIORITY) {
		disable_ipc_src(VPE0_MIPS_TIMER_INT);
		do_IRQ(VPE0_MIPS_TIMER_INT);
		return;
	}

	/* Process interrupts on other interrupt lines */
	stat = ipc->status_low[prio+1];
	if(likely(stat != 0))
        {
	  /* Very first thing to check if there is some thing to do */
	  while (stat) 
	  {
	 	irq = ffs(stat) - 1;
		stat &= ~(1 << irq);
		IRQSet1[IRQCnt1++] = irq;
          }
        }

	stat = ipc->status_hi[prio + 1];

        if(likely(stat != 0))
        {
          /* Very first thing to check if there is some thing to do */
          while (stat)
          {
                irq = ffs(stat) - 1;
                stat &= ~(1 << irq);
                IRQSet2[IRQCnt2++] = irq + 32;
          }
        }
        
	  /* 
	     We should not call do_IRQ before we acknowldege the 
             Interupt Controller , one or more IRQ handler may not 
             have SA_INTERRUPT bit set and kernel will enable the 
             interrupts (which was set by processor ), which will 
             again drag some body here . 
	  */

	  /* Let us acknowldege Interupt Controller that we got the message */
	  for(i = 0; i<IRQCnt1; i++)
  	  {
            disable_ipc_src(IRQSet1[i]);
          }

	  /* Let us acknowldege Interupt Controller that we got the message */
	  for(i = 0; i<IRQCnt2; i++)
  	  {
            disable_ipc_src(IRQSet2[i]);
          }
	
	  /* 
	     Now we are safe to enable the interrupts becauase we have 
	     told Interupt Controller that we are all set to go. 
          */
	  for(i = 0; i<IRQCnt1; i++)
          {
	    do_IRQ(IRQSet1[i]);
	  }

	  /* 
             Now we are safe to enable the interrupts becauase we have 
	     told Interupt Controller that we are all set to go. 
          */
	  for(i = 0; i<IRQCnt2; i++)
          {
	    do_IRQ(IRQSet2[i]);
	  }
}

#if defined (CONFIG_CPU_MIPSR2_IRQ_VI)
void mipscpu_int_hw5_dispatch(void)
{        
        do_IRQ(MIPSCPU_INT_HW5);
}

void mipscpu_int_hw4_dispatch(void)
{
	do_IRQ(MIPSCPU_INT_HW4);
}

void mipscpu_int_hw3_dispatch(void)
{
        do_IRQ(MIPSCPU_INT_HW3);
}

void mipscpu_int_hw2_dispatch(void)
{
	do_IRQ(MIPSCPU_INT_HW2);
}

void mipscpu_int_hw1_dispatch(void)
{
	do_IRQ(MIPSCPU_INT_HW1);
}

void mipscpu_int_hw0_dispatch(void)
{
	do_IRQ(MIPSCPU_INT_HW0);
}

unsigned int status_ipc_src(unsigned int irq_nr)
{
	unsigned char ipc_reg_select = irq_nr/8, ipc_src = irq_nr % 8;
	unsigned int dest_line;
	unsigned long stat;

	/* get the status of the source */
	dest_line = (IPC_DEFAULT_DEST_LINE << (ipc_src * 4));
	/* get the destination interrupt line */
	dest_line = (ipc->dest_select[ipc_reg_select] & dest_line) >> (ipc_src * 4);

	if(irq_nr < 32)	{
		stat = ipc->status_low[dest_line];
		return (stat & ( 1 << irq_nr) ? 1 : 0);
	}
	else {
		stat = ipc->status_hi[dest_line];
		return (stat & ( 1 << irq_nr - 32) ? 1 : 0);
	}
}

EXPORT_SYMBOL(status_ipc_src);

void setup_vint_handlers (void) {
	if (cpu_has_vint) {
		//The despatch for h/w int 5 is defined in plat_time.c file
		set_vi_handler(MIPSCPU_INT_HW4, mipscpu_int_hw4_dispatch);
		set_vi_handler(MIPSCPU_INT_HW3, mipscpu_int_hw3_dispatch);
		set_vi_handler(MIPSCPU_INT_HW2, mipscpu_int_hw2_dispatch);
		set_vi_handler(MIPSCPU_INT_HW1, mipscpu_int_hw1_dispatch);
		set_vi_handler(MIPSCPU_INT_HW0, mipscpu_int_hw0_dispatch);
		startup_ipc_src(UART1_INT);
	}
}
#endif //CONFIG_CPU_MIPSR2_IRQ_VI

void __init fusiv_init_IRQ(void)
{
	int i;
	unsigned long cp0_status;

	cp0_status = read_c0_status();

	printk("Initializing IPC..\n");

	for (i = 0; i < 8; i++)
		ipc->src_ctrl[i] = 0x0;

	/* Route all sources by default to IPC dest line 15 which 
           is unconncted */
	for (i = 0; i < 8; i++)
		ipc->dest_select[i] = 0xFFFFFFFF;
	
	/* Clear all status registers */
	for (i = 0; i < 15; i++)
		ipc->status_low[i] = 0xFFFFFFFF;
	for (i = 0; i < 15; i++)
		ipc->status_hi[i] = 0xFFFFFFFF;

#if defined (CONFIG_CPU_MIPSR2_IRQ_VI)
	/* Initialize IRQ descriptor for six MIPS interrupt lines (2-7) */
	if (!cpu_has_veic && cpu_has_vint) 
	{
		mips_cpu_irq_init();
		setup_vint_handlers();
	} 
	else
#endif
	{
		/* let us initialize all the IRQs Descriptors. */
		for (i = 0; i < VX185_MAX_INTS; i++)
		{
			irq_desc[i].chip     = &vx185_ipc;
			irq_desc[i].status	= IRQ_DISABLED;
			irq_desc[i].action	= 0;
			irq_desc[i].depth	= 1;
			spin_lock_init(&irq_desc[i].lock);
		}

		clear_c0_status(ALL_INTS);
		set_c0_status(ALL_INTS);
	}

#ifdef CONFIG_KGDB
	/* If local serial I/O used for debug port, enter kgdb at once */
	puts("Waiting for kgdb to connect...");
	set_debug_traps();
	breakpoint();
#endif
	
	
}

void unmask_mips_timer_irq(unsigned int irq)
{
    if(irq == MIPS_TIMER_INT){
        enable_ipc_src(irq);
    } else {
        printk(KERN_ERR "[%s] ignoring request, irq %d != MIPS_TIMER_INT\n", __func__, irq);
    }
}

void mask_mips_timer_irq(unsigned int irq)
{
    if(irq == MIPS_TIMER_INT){
        disable_ipc_src(irq);
    } else {
        printk(KERN_ERR "[%s] ignoring request, irq %d != MIPS_TIMER_INT\n", __func__, irq);
    }
}
