/* Copyright (C) 2006 Ikanos Communications
 * See IKANOS_PROP_LICENSE.txt for license information. */

// ===========================================================================
//
// File name    : $RCSfile: aphw.h,v $
// Description  : This file defines the I/O port addresses and bit allocations
//                for all AP hardware used in the Vox200/Secura 1xx devices
// ===========================================================================

#ifndef _APHW_H_
#define _APHW_H_

/* This file provides brings together all the files which define the I/O
 * ports available on Vox 200 / Secura 110 / Secura 120 AP implementations.
 *
 * To simplify defining the base addresses for peripherals, especially
 * in some cases where the are several instances of some peripherals, the
 * I/O register addresses are "calculated" by macros which include
 * instance and channel numbers where apprporiate.
 *
 * For example:
 * The main AP control register is specified by _AP_ADDR(CTRL)
 * The control register for AHB DMA channel 2 is _AP_ADDR(2, CRTL)
 * The control register for input queue 1 is _INPQ_ADDR(1, CTRL)
 *
 * Input flag numbers are defined the same way
 *
 * The description of each block gives the valid parameters for each macro
 */


/* Useful macro to convert bit numbers to values */
#define _EXP(x) (1 << (x ## _BIT))

/* Include the definitions for the basic I/O block. All APs have this hardware
 *
 * Valid reg names are CTRL, PC, INTERRUPT_STATUS, INTERRUPT_MASK,
 * GP_REGISTER_1 to GP_REGISTER_6, EVENT_IN, EVENT_OUT, TMR_VALUE,
 * TMR_LIMIT, TMR_CTRL, TMR_STATUS
 *
 * Valid flag names are TIMER and INTERRUPT
 */
#include <netpro/apinternal.h>

#define AP_TIMER_FLAG           0       // Set to indicate timer overflow
#define AP_INTERRUPT_FLAG       1       // Interupt input from ICU to AP

#define _AP_ADDR(reg)   ( AP_ ## reg )
#define _AP_FLAG(flg)   ( AP_ ## flg ## _FLAG)

/* Include definitions for the AHB DMA controller. Each AP has a single
 * instance of this hardware.
 *
 * The AHB DMA controller supports 4 seperate channels.
 *
 * Valid chan values are 0 .. 3
 * Valid reg names are CTRL, STATUS, MSW, LSW, DMEM, LENGTH
 *
 * Flags 0 .. 3 are associated with channels 0 .. 3
 *
 * Note that the addresses of AHB DMA channel registers are greater than
 * 0x1f and cannot be accessed directly
 */
#include <netpro/apahbdma.h>

#define _AHB_ADDR(chan, reg)  \
        (((chan) * AP_AHB_CHAN_LEN) + AP_AHB_ ## reg ## _OFF + 0x20)
#define _AHB_FLAG(chan)         ((chan) + 4)

///ATTENTION !!! Buffer Manager APU has reverse direction of interrupt bits inside Flag register:
// Bit 4 - for DMA channel 3, 5 - 2, 6 - 1, 7 - 0
//No need for common Vox150/Vox200 code #define _AHB_FLAG_BM(chan)      (7 - (chan))

/* Include definitions for the input queue block. All APs have at least
 * a single input queue. The buffer manager AP has three!
 *
 * Valid queue numbers are 1 .. 3
 * Valid reg names are CTRL, STATUS, MSW, LSW
 *
 * There is a single non-empty flag bit associated with each queue
 */
#include <netpro/apinpq.h>

#define _INPQ_ADDR(queue, reg)  \
        ((((queue) - 1) * 4) + AP_INPQ_ ## reg ## _OFF + 0x10)
#define _INPQ_FLAG(queue) (((queue) - 1) + 8)

/* Include definitions for the output queue block. The buffer management
 * AP has two output queues. Other APs do not have any output queues
 *
 * Valid queue numbers are 1 or 2
 * Valid reg names are CTRL, STATUS, MSW, LSW
 *
 * There is a single non-full flag bit associated with each queue
 *
 * Note that output queue addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include <netpro/apoutpq.h>

#define _OUTPQ_ADDR(queue, reg)  \
        ((((queue) - 1) * 4) + AP_OUTPQ_ ## reg ## _OFF + 0x40)
#define _OUTPQ_FLAG(queue) (((queue) - 1) + 11)

/* Include definitions for the ethernet DMA controller block. All ethernet
 * APs have a single instance of this block.
 *
 * Valid reg names are CTRL, RX_MSW, RX_LSW, TX_CTRL, RX_ADDR, RX_LENGTH,
 * TX_ADDR, TX_LENGTH
 *
 * Valid flg values are TX_ACTIVE and RX_ACTIVE
 */
#include <netpro/apeth.h>

#define AP_ETH_RX_ACTIVE_FLAG   10
#define AP_ETH_TX_ACTIVE_FLAG   11

#define _ETH_ADDR(reg) ((AP_ETH_ ## reg ## _OFF) + 0x18)
#define _ETH_FLAG(flg) (AP_ETH_ ## flg ## _FLAG)

/* Include definitions for the Utopia DMA block. This block is only present
 * on the third ethernet AP, which does also handles UTOPIA traffic (The
 * Utopia and ethernet 3 functions are mutually exclusive)
 *
 * Valid reg names are RX_CTRL, RX_ADDR, RX_STATUS, TX_CTRL, TX_ADDR, VCC_CFG
 * Valid flag names are RX_ACTIVE, TX_ACTIVE
 *
 * Note that utopia dma register addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include <netpro/aputopia.h>

#define AP_UTP_RX_ACTIVE_FLAG   10
#define AP_UTP_TX_ACTIVE_FLAG   11

#define _UTOPIA_ADDR(reg) ((AP_UTP_ ## reg ## _OFF) + 0x50)
#define _UTOPIA_FLAG(flg) (AP_UTP_ ## flg ## _FLAG)

/* Include definitions for the Security DMA and CSRs. These are only
 * present in the security AP
 *
 * Valid reg names are RX_CTRL, RX_ADDR, RX_LENGTH, TX_CTRL, TX_ADDR, TX_LEN
 * ENCR_CTRL_LSW, ENCR_CTRL_MSW, AUTH_CTRL, HASH_BASE, KEY_BASE, IV_BASE,
 * NEXT_IV_BASE, MSGLEN_BASE, KEY_HASH_BASE, IV_HASH_BASE
 *
 * The reg names ending in "_BASE" give the first address in a block of
 * registers associated with the function. Typically these will be filled
 * in a loop and are not individually addressed.
 *
 * Valid flg names are CEN_BUSY, SPA_RX, and SPA_TX0
 *
 * Note that security register addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include <netpro/apsecurity.h>

#define AP_SEC_CEN_BUSY_FLAG    9
#define AP_SEC_RX_ACTIVE_FLAG   10
#define AP_SEC_TX_ACTIVE_FLAG   11

#define _SEC_ADDR(reg) ((AP_SEC_ ## reg ## _OFF) + 0x60)
#define _SEC_FLAG(flg) (AP_SEC_ ## flg ## _FLAG)

#endif /* _APHW_H_ */
