/* =====================================================================
 * Copyright (C) 2006 Ikanos Communications 
 * See IKANOS_PROP_LICENSE.txt for license information.
 * =====================================================================
 */

/* ===========================================================================
 * File name    : $RCSfile: apvdsl.h,v $
 * Description  : The I/O layout and bit definitions for the AP VDSL
 *                hardware and data source simulation
 * ===========================================================================
 */

#ifndef _APVDSL_H_
#define _APVDSL_H_

/* This is in some ways best thought of as a combination on the Vx200-type
 * ethernet and ATM controllers as it operates in both a cell-based mode and
 * a packet based mode.
 *
 * Wherever possible the same definitions are used as the pre-existing
 * devices, but Benny (the VLSI engineer) has decided to do some restructuring
 * that will require software changes.
 *
 * Registers that start with 'VDSL' are used in both packet and cell mode,
 * registers starting with 'ATM' are only used in cell mode and registers
 * starting with 'PTM' are only used in packet mode.
 *
 * Also note that the VDSL PHY supports four bearer channels. These act
 * independently. As there is a single Tx and Rx DMA controller it is
 * important that the AP firmware avoids initiating any DMA operation the
 * would cause the DMA controller to block due to a full (TX) or empty (RX)
 * FIFO. The bearer for the operation is set in the control register.
 *
 * Finally the bearer FIFOs have some strange habits. These are best explained
 * by the microarchitecture documents, but in summary:
 *
 * RX: A RxDMA operation should not end with less than two slot buffers of
 * data in the bearer slot buffer unless there is an EOP in the FIFO. For
 * example, for a 128 byte RxDMA there should be at least 4 (* 64 bytes) slot
 * buffers containing RxDMA unless there is an EOP in the FIFO.
 *
 * TX: The bearer Tx slot buffer underflows if they contain less than a
 * configurable amount of data or an EOP (unfortunately it cannot be
 * configured for zero threshold). A typical threshold is 2 - 3 (*64 bytes)
 * slots - this means that the TxDMA size should be at least 256 bytes so that
 * the TX side doesn't quickly underrun.
 */

/* The VDSL_TX_CTRL register controls the operation of the data transmit side
 * of the peripheral. The bearer is specified by the TX_QUEUE entries
 */
#define VDSL_TX_CTRL_OFF        0x0
#define VDSL_TX_START_BIT       0       // Set to start Tx DMA
#define VDSL_TX_ABORT_BIT       1       // Abort Tx DMA (not simulated)
#define VDSL_TX_QUEUE_MASK      0x000c  // Queue number in PTM (pkt) mode
#define VDSL_TX_QUEUE_SHFT      2
#define VDSL_TX_GFC_MASK        0x00f0  // Cell header GFC field
#define VDSL_TX_GFC_SHFT        4
#define VDSL_TX_PTI_MASK        0x0700  // Cell header PTI field
#define VDSL_TX_PTI_SHFT        8
#define VDSL_TX_CLP_BIT         11      // Cell header CLP bit
#define VDSL_TX_PTM_BIT         15      // Set transfer mode (1 = PTM [pkt])

/* The VDSL_TX_ADDR register holds the base address of the data being
 * transmitted.
 */
#define VDSL_TX_ADDR_OFF        0x2

/* The ATM_TX_CHAN_CTRL register controls how ATM cells are transmitted. This
 * must be correctly configured before initiating the DMA that transfers the
 * cell data
 */
#define ATM_TX_CHAN_CTRL_OFF    0x4
#define ATM_TX_CHAN_MASK        0x001f   // ATM channel number
#define ATM_TX_CHAN_SHFT        0
#define ATM_TX_HDR_EN_BIT       6        // Enable header insertion
#define ATM_TX_PHY_MASK         0x0f80   // PHY (bearer) number
#define ATM_TX_PHY_SHFT         7
#define ATM_TX_CRC_EN_BIT       12       // Enable CRC insertion in last cell of
                                         // AAL5 frames
#define ATM_TX_LAST_BIT         13       // Last cell in AAL5 frame
#define ATM_TX_UNMAP_CHAN_BIT   14       // ANIL TBD
#define ATM_TX_UNMAP_HDR_BIT    15       // ANIL TBD

/* The PTM_TX_LEN register holds the length, CRC, and last segment
 * information in packet mode operation.
 *
 * Note that this isn't used in cell mode operation. In that mode there
 * is no need to denote the last DMA (all the data must be in a single DMA)
 * and the length is implicit in the type of DMA operation requested
 */
#define PTM_TX_LENGTH_OFF       0x6
#define PTM_TX_LENGTH_MASK      0x3fff   // Packet segment length
#define PTM_TX_LENGTH_SHFT      0
#define PTM_TX_CRC_EN_BIT       14       // Enable append of packet CRC
#define PTM_TX_LAST_BIT         15       // Indicates last segment of packet

/* The VDSL_TX_SLB() registers give information about the state of the
 * bearer transmit queues. This is necessary to prevent blocking DMA
 * operations (as noted above).
 *
 * These is one of these registers for each bearer
 */
#define _VDSL_TX_SLB_OFF(n) (2 * (n) + 0x8)
#define VDSL_TX_SLB_AVAIL_MASK  0x3fff   // Number of bytes left in queue
#define VDSL_TX_SLB_AVAIL_SHFT  0
#define VDSL_TX_SLB_URUN_BIT    14       // Underrun has occurred for the queue

/* The VDSL_RX_CTRL register contains control bits related to VDSL receive
 * (downstream traffic) operation. Like the VDSL_TX_CTRL the current bearer
 * is specified in htis register.
 *
 * Note that there is no need for the  TxDMA and RxDMA queue (bearer) need to
 * be the same.
 */
#define VDSL_RX_CTRL_OFF        0x10
#define VDSL_RX_START_BIT       0        // Start Rx DMA
#define VDSL_RX_ABORT_BIT       1        // Abort Rx DMA (not simulated)
#define VDSL_RX_QUEUE_MASK      0x000c   // Queue number for Rx
#define VDSL_RX_QUEUE_SHFT      2
#define VDSL_RX_ERR_BIT         14       // PHY error in packet (only at EOP)
#define VDSL_RX_PTM_BIT         15       // Receive transfer mode (1 = PTM [pkt])

/* The VDSL_RX_ADDR reigster holds the address in DMEM of where to put the
 * next cell or packet (depending on the receiver mode)
 */
#define VDSL_RX_ADDR_OFF        0x12

/* The ATM_RX_CHAN_STS register holds information about the last cell
 * received
 */
#define ATM_RX_CHAN_STS_OFF     0x14
#define ATM_RX_VCC_BIT          0        // Incoming cell matched header entry
#define ATM_RX_PTI_MASK         0x000e   // PTI bits of ATM header
#define ATM_RX_PTI_SHFT         0
#define ATM_RX_PERR_BIT         4        // Cell parity error (?)
#define ATM_RX_HEC_ERR_BIT      5        // Cell HEC error
#define ATM_RX_PHY_MASK         0x07c0   // PHY number (= queue/bearer number)
#define ATM_RX_PHY_SHFT         6

/* The ATM_RX_STS register holds more information about the last cell. Some
 * of this information is only valid if the ATM_RX_VCC_BIT is set in
 * ATM_RX_CHAN_STS
 */
#define ATM_RX_STS_OFF         0x16
#define ATM_RX_VCC_MASK        0x003f   // Header entry that matched VCC
#define ATM_RX_CRC_ERR_BIT     6        // CRC error in AAL5 frame
#define ATM_RX_CLP_BIT         7        // CLP bit of cell
#define ATM_RX_GFC_MASK        0x0f00   // GFC field from cell

/* The PTM_RX_LENGTH register specifies the length of the DMEM buffer
 * before RXDMA in packet mode, and is set to the actual segment length
 * at the end of the DMA (in packet mode)
 */
#define PTM_RX_LENGTH_OFF       0x18
#define PTM_RX_LENGTH_MASK      0x3fff    // Wr: Max RX len, Rd: Actual Rx len
#define PTM_RX_LENGTH_SHFT      0
#define PTM_RX_ERR_BIT          14        // Error in received packet (if EOP=1)
#define PTM_RX_EOP_BIT          15        // End of frame received

/* The VDSL_RX_SLB() registers give information about the state of the
 * per-bearer receive queues. There is one of these registers for each
 * bearer queue
 */
#define _VDSL_RX_SLB_OFF(n) (2 * (n) + 0x1a)
#define VDSL_RX_SLOT_CNT_MASK   0x01ff    // Number of slots in queue
#define VDSL_RX_ORUN_BIT        14        // An error (overrun) has occured
                                          // on this queue (W1C)
#define VDSL_RX_EOP_BIT         15        // At least one EOP in buffer

/* The VDSL_RX_EOP registers hold the number of EOP's in each slot buffer.
 * Like the VDSL_RX_SLB register these is one of these for each queue
 */
#define _VDSL_RX_EOP_OFF(n) (2 * (n) + 0x22)
#define VDSL_RX_EOP_MASK        0x01ff    // Number of EOP's in bearer slot buf
#define VDSL_RX_EOP_SHFT        0

#endif /* _APVDSL_H_ */
