/* =====================================================================
 * Copyright (C) 2006 Ikanos Communications
 * See IKANOS_PROP_LICENSE.txt for license information.
 * =====================================================================
 */

/* ===========================================================================
 * File name    : $RCSfile: apvx185hw.h,v $
 * Description  : This file defines the I/O port addresses and bit allocations
 *                for all AP hardware used in the Vx185 devices
 * ===========================================================================
 */

#ifndef _APVX185HW_H_
#define _APVX185HW_H_

/* This file provides brings together all the files which define the I/O
 * ports available on Vx185
 *
 * To simplify defining the base addresses for peripherals, especially
 * in some cases where the are several instances of some peripherals, the
 * I/O register addresses are "calculated" by macros which include
 * instance and channel numbers where apprporiate.
 *
 * For example:
 * The main AP control register is specified by _AP_ADDR(CTRL)
 * The control register for AHB DMA channel 2 is _AHB_ADDR(2, CRTL)
 * The control register for input queue 1 is _INPQ_ADDR(1, CTRL)
 *
 * Input flag numbers are defined the same way
 *
 * The description of each block gives the valid parameters for each macro
 */


/* Useful macro to convert bit numbers to values */
#define _EXP(x) (1 << (x ## _BIT))

/* Include the definitions for the basic I/O block. All APs have this hardware
 *
 * Valid reg names are CTRL, PC, INTERRUPT_STATUS, INTERRUPT_MASK,
 * GP_REGISTER_1 to GP_REGISTER_6, EVENT_IN, EVENT_OUT, TMR_VALUE,
 * TMR_LIMIT, TMR_CTRL, TMR_STATUS
 *
 * Valid flag names are TIMER and INTERRUPT
 */
#include "../apinternal.h"

#define AP_TIMER_FLAG           0       // Set to indicate timer overflow
#define AP_INTERRUPT_FLAG       1       // Interupt input from ICU to AP

#define _AP_ADDR(reg)   ( AP_ ## reg )
#define _AP_FLAG(flg)   ( AP_ ## flg ## _FLAG)

/* Include definitions for the AHB DMA controller. Each AP has a single
 * instance of this hardware.
 *
 * The AHB DMA controller supports 4 seperate channels.
 *
 * Valid chan values are 0 .. 3
 * Valid reg names are CTRL, STATUS, MSW, LSW, DMEM, LENGTH
 *
 * Flags 0 .. 3 are associated with channels 0 .. 3
 *
 * Note that the addresses of AHB DMA channel registers are greater than
 * 0x1f and cannot be accessed directly
 */
#include "../apahbdma.h"

#define _AHB_ADDR(chan, reg)  \
        (((chan) * AP_AHB_CHAN_LEN) + AP_AHB_ ## reg ## _OFF + 0x20)
#define _AHB_FLAG(chan)         ((chan) + 2)

#define _AHB_DBUS_ADDR(chan, reg)    _AHB_ADDR(chan, reg)


#define _AHB_CBUS_ADDR(chan, reg)       \
        (((chan) * AP_AHB_CHAN_LEN) + AP_AHB_ ## reg ## _OFF + 0xE0)

/* Include definitions for the input queue block. All APs have at least
 * a single input queue. The buffer manager AP has three!
 *
 * Valid queue numbers are 1 .. 3
 * Valid reg names are CTRL, STATUS, MSW, LSW
 *
 * There is a single non-empty flag bit associated with each queue
 */
#include "../apinpq.h"

#define _INPQ_ADDR(queue, reg)  \
        ((((queue) - 1) * 4) + AP_INPQ_ ## reg ## _OFF + 0x10)
#define _INPQ_FLAG(queue) (((queue) - 1) + 6)

/* Include definitions for the output queue block. The buffer management
 * AP has two output queues. Other APs do not have any output queues
 *
 * Valid queue numbers are 1 or 2
 * Valid reg names are CTRL, STATUS, MSW, LSW
 *
 * There is a single non-full flag bit associated with each queue
 *
 * Note that output queue addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include "../apoutpq.h"

#define _OUTPQ_ADDR(queue, reg)  \
        ((((queue) - 1) * 4) + AP_OUTPQ_ ## reg ## _OFF + 0x40)
#define _OUTPQ_FLAG(queue) (((queue) - 1) + 11)

/* Include definitions for the ethernet DMA controller block. All ethernet
 * APs have a single instance of this block.
 *
 * Valid reg names are CTRL, RX_STATUS, RX_ADDR, RX_LENGTH, TX_ADDR, TX_LENGTH
 *
 * Valid flag values are TX_ACTIVE and RX_ACTIVE
 */
#include "../apgige.h"

//////////#define AP_ETH_RX_ACTIVE_FLAG   10
//////////#define AP_ETH_TX_ACTIVE_FLAG   11
// Reversed in Vox180
#define AP_ETH_RX_ACTIVE_FLAG   11
#define AP_ETH_TX_ACTIVE_FLAG   10

#define _ETH_ADDR(reg) ((AP_ETH_ ## reg ## _OFF) + 0x18)
#define _ETH_FLAG(flg) (AP_ETH_ ## flg ## _FLAG)

/* Move the following later */
#define GIGE_TX_CONTROL       	0x180C
#define GIGE_TX_SDRAM       	0x180D
#define GIGE_TX_DMEM       	0x180E
#define GIGE_TX_LENGTH       	0x180F

/* Include definitions for the VDSL (Cell and Packet Mode) block. The
 * registers are divided into three groups VDSL, ATM, and PTM (depending
 * on which mode of operation they are used in)
 *
 * Valid VDSL names are TX_CTRL, TX_ADDR, TX_SLB(*), RX_CTRL, RX_ADDR,
 * RX_SLB(*), RX_EOP(*)
 * Valid PTM names are TX_LENGTH, RX_LENGTH,
 * Valid ATM names are TX_CHAN_CTRL, RX_CHAN_STS
 * Valid flag names are RX_ACTIVE, TX_ACTIVE
 *
 * Note that VDSL register addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include "../apvdsl.h"
#include "../aputopia.h"

#define AP_VDSL_RX_ACTIVE_FLAG   10
#define AP_VDSL_TX_ACTIVE_FLAG   11

#define VDSL_BASE               0x2050

//#define _VDSL_ADDR(reg) ((VDSL_ ## reg ## _OFF) + VDSL_BASE)
#define _VDSL_ADDR(reg) (( VDSL_ ## reg ## _OFF) + VDSL_BASE)
#define _VDSL_ADDR_BR(bearer, reg) (_VDSL_ ## reg ## _OFF(bearer) + VDSL_BASE)
#define _ATM_ADDR(reg) ((ATM_ ## reg ## _OFF) + VDSL_BASE)
#define _PTM_ADDR(reg) ((PTM_ ## reg ## _OFF) + VDSL_BASE)
#define _VDSL_FLAG(flg) (AP_VDSL_ ## flg ## _FLAG)

/* Include definitions for the Security DMA and CSRs. These are only
 * present in the security AP
 *
 * Valid reg names are RX_CTRL, RX_ADDR, RX_LENGTH, TX_CTRL, TX_ADDR, TX_LEN
 * ENCR_CTRL_LSW, ENCR_CTRL_MSW, AUTH_CTRL, HASH_BASE, KEY_BASE, IV_BASE,
 * NEXT_IV_BASE, MSGLEN_BASE, KEY_HASH_BASE, IV_HASH_BASE
 *
 * The reg names ending in "_BASE" give the first address in a block of
 * registers associated with the function. Typically these will be filled
 * in a loop and are not individually addressed.
 *
 * Valid flg names are CEN_BUSY, SPA_RX, and SPA_TX0
 *
 * Note that security register addresses are greater than 0x1f and
 * cannot be accessed directly
 */
#include "../apsecurity.h"

#define AP_SEC_CEN_BUSY_FLAG    11 //9
#define AP_SEC_RX_ACTIVE_FLAG   2  //10
#define AP_SEC_TX_ACTIVE_FLAG   3  //11

#define _SEC_ADDR(reg) ((AP_SEC_ ## reg ## _OFF) + 0x60)
#define _SEC_FLAG(flg) (AP_SEC_ ## flg ## _FLAG)

#define DMA_START_BIT           0
#define CEN_RX_CTRL_ADDR        0x64
#define CEN_TX_CTRL_ADDR        0x60
#define WAIT_CEN_RX_DMA(reg, label)    lbl     reg, CEN_RX_CTRL_ADDR \
                                       in      reg, reg              \
					                   nop                           \
									   if  bs  reg, DMA_START_BIT  jmp  label

#define WAIT_CEN_TX_DMA(reg, label)    lbl     reg, CEN_TX_CTRL_ADDR \
                                       in      reg, reg              \
					                   nop                           \
									   if  bs  reg, DMA_START_BIT  jmp  label

#endif /* _APHW_H_ */
